Sets
    i Index of buses /1*2/
    l Index of lines /1/

    k Index of linearization segments of voltage magnitude term /1*4/
    s Index of linearization segments of circular constraint /1*30/
    
    t Index of time /1*5/
    a Index of agents /1*5/


*   ********************
    iterations /1*10/
    itEndSet(iterations) /10/
*   ********************


;

*need two symbols to references buses for double iterations
alias(i,j); 
*tEnd(t) = t$

Scalars

    iEnd
    tEnd
    aEnd
    itEnd

    v_min /0.9/
    v_ref /1.0/
    v_max /1.05/

    theta_min /-3.14/
    theta_max /3.14/

    PG_min /0/
    PG_max /25/
    QG_min /-1/
    QG_max /1/

    SOC_min /0.2/

    BL_limit /0.5/

    M /999999999/

*    nG marginal price of G /5/

    dalfa Angle deviation
    dV_min /0/
    dV_max Maximum voltage deviation

    iteration
    iteration_limit
;

aEnd = card(a);
tEnd = card(t);
iEnd = card(i);
itEnd = card(iterations);

iteration_limit = card(iterations);

dalfa = 2*pi/30;
dV_max = (v_max - v_min) / 4;


Table Line(l,*) Line characteristics

    se  re  R       X       Smax
1   1   2   0.002   0.001   25
;

Parameter is_connected_to(a,i,t);
*            1   2   3   4   5
*    1.2     1   1   0   0   1
 
*    2.2     1   1   0   0   1

*    3.2     1   1   0   0   1
*;

is_connected_to(a,i,t)$((ord(t) = 1 or ord(t) = 2 or ord(t) = 5) and ord(i) = 2) = 1;


Table Pbld(i, t)

    1       2       3       4       5
1   0       0       0       0       0
2   0.2     0.5     0.2     0.5     0.2

;

Table Qbld(i, t)

    1       2       3       4       5
1   0.05    0.07    0.04    0.07    0.04
2   0.01    0.03    0.02    0.03    0.02

;

Parameter price_of_power(i,t);
price_of_power(i,t) = 8;
price_of_power(i,t)$(ord(t)=1) = 10;
price_of_power(i,t)$(ord(t)=2 or ord(t)=4) = 14;

Parameters
 
    GL(l)           Conductance of a line in fundamental frequency
    BL(l)           Susceptance of a line in fundamental frequency

    MG(i,j)         Conductance matrix in fundamental frequency
    MB(i,j)         Susceptance matrix in fundamental frequency

    A1(l,i,j)
    A2(l,i,j)
    AL(i,j)

    SLmax(i,j)
    hh(s)

    initial_SOC(a)
    charging_rate(a)
    discharge_rate(a)
    capacity(a)

    nG(t)

    ich_1(t)
    ich_2(t)

    lambda(t, iterations)
    mew_3(t, iterations)
    marginals(i,t, iterations)

;

GL(l)=Line(l,'R')/(sqr(Line(l,'R'))+sqr(Line(l,'X')));
BL(l)=-Line(l,'X')/(sqr(Line(l,'R'))+sqr(Line(l,'X')));


A1(l,i,j) $ (ord(i) = Line(l,'se') and ord(j) = Line(l,'re')) = 1;
A2(l,i,j) $ (ord(j) = Line(l,'se') and ord(i) = Line(l,'re')) = 1;
AL(i,j) = sum(l, A1(l,i,j) + A2(l,i,j) );

MG(i,j)=sum(l, GL(l)*(A1(l,i,j)+A2(l,i,j)));
MB(i,j)=sum(l, BL(l)*(A1(l,i,j)+A2(l,i,j)));

SLmax(i,j) = sum(l, Line(l,'Smax') * (A1(l,i,j) + A2(l,i,j)) );


hh(s) = ord(s) * dalfa;

initial_SOC(a) = 0.5;
initial_SOC(a)$(ord(a) = 1)    = 0.45;
initial_SOC(a)$(ord(a) = 2)    = 0.9;
initial_SOC(a)$(ord(a) = 3)    = 0.5;


charging_rate(a)  = 1.0;
discharge_rate(a) = 0.6;
capacity(a)       = 10.0;

nG(t) = 4;
nG(t)$(ord(t) = 2 or ord(t) = 4) = 7;
*nG(t)$(ord(t) = 1) = 3;

Free variables

    z_c
    z_dc
    
    Pg(i,t)
    Qg(i,t)

    theta(i,t)

    SOL

    PL(i,j,t)
    QL(i,j,t)

    b(a, t)
    tu

    w(a,t)

    P_34
    P_34_time(a, t)
    
    P_35
    P_35_time(t)
;

Positive variable

    is_charging(a, t)

;
is_charging.up(a,t) = 1.0;

Binary variable
    bw(a,t)
;

Positive variables
    
    PCh(a, t)
    PCo(a, t)

    SOC(a, t)

    Pald(i,t)
    Qald(i,t)

    dV(i,t,k)
    voltage(i,t)
;

* Non generator buses don't have Pg or Qg
Pg.fx(i,t)$(ord(i) > 1) = 0;
Qg.fx(i,t)$(ord(i) > 1) = 0;

Pg.lo(i,t)$(ord(i) = 1) = PG_min;
Pg.up(i,t)$(ord(i) = 1) = PG_max;

Qg.lo(i,t)$(ord(i) = 1) = QG_min;
Qg.up(i,t)$(ord(i) = 1) = QG_max;

theta.fx(i,t)$(ord(i) = 1) = 0;

theta.lo(i,t)$(ord(i) > 1) = theta_min;
theta.up(i,t)$(ord(i) > 1) = theta_max;

dV.lo(i,t,k) = dV_min;
dV.up(i,t,k) = dV_max;

dV.fx(i,t,k)$(ord(i) = 1 and ord(k) = 1) = dV_max;
dV.fx(i,t,k)$(ord(i) = 1 and ord(k) = 2) = dV_max;
dV.fx(i,t,k)$(ord(i) = 1 and ord(k) = 3) = v_ref - v_min - 2 * dV_max;
dV.fx(i,t,k)$(ord(i) = 1 and ord(k) = 4) = 0;

Equations

    Z_central_
    Z_central_dumb_

    HPF1_(i,t)
    HPF2_(i,t)

    HPF3_l_(i,j,t)
    HPF4_l_(i,j,t)

    HPF3_nl_(i,j,t)
    HPF4_nl_(i,j,t)

    SOL_1_(i,j,t,s)
    SOL_2_(i,j,t,s)
    SOL_v_(i,j,t,s)

    Voltage_(i,t)


    Z_decentral_

    Pald_(i,t)
    Qald_(i,t)

    SOC_initial_(a,t)
    SOC_(a,t)

    SOC_min_(a,t)
    SOC_max_(a,t)

    eq_soc_(a, t)

    PCh_(a,t)
    PCo_(a,t)

    eq_34_
    eq_34_t_(a, t, iterations)
    eq_34_t_balance_(a, t,iterations)

    eq_35_
    eq_35_t_(t, iterations)
    eq_35_t_balance_(t, iterations)
;

Z_central_.. z_c =e= sum((i,t), nG(t) * Pg(i,t));
Z_central_dumb_.. z_c =e= sum((a,t) , SOC(a, t) * (2- ord(t)/card(t)));

HPF1_(i,t)..    Pg(i,t) - Pbld(i,t) - Pald(i,t) =e= sum(j, AL(i,j) * PL(i,j,t));
HPF2_(i,t)..    Qg(i,t) - Qbld(i,t) - Qald(i,t) =e= sum(j, AL(i,j) * QL(i,j,t));

HPF3_l_(i,j,t)..   PL(i,j,t) =e= MG(i,j) * v_ref * sum(k, dV(i,t,k) - dV(j,t,k)) - MB(i,j) * sqr(v_ref) * (theta(i,t) - theta(j,t));
HPF4_l_(i,j,t)..   QL(i,j,t) =e= MB(i,j) * v_ref * sum(k, dV(i,t,k) - dV(j,t,k)) - MG(i,j) * sqr(v_ref) * (theta(i,t) - theta(j,t));

HPF3_nl_(i,j,t)..   PL(i,j,t) =e=  voltage(i,t) * voltage(j,t) * MG(i,j) * cos( theta(i,t) - theta(j,t) - MB(i, j));
HPF4_nl_(i,j,t).. QL(i,j,t) =e= - (voltage(i,t) * voltage(j,t) * MG(i,j) * sin(-theta(i,t) + theta(j,t) + MB(i, j)) );

SOL_1_(i,j,t,s).. SOL(i,j,t,s) =l= SLmax(i,j);
SOL_2_(i,j,t,s).. SOL(i,j,t,s) =g= -SLmax(i,j);
SOL_v_(i,j,t,s).. SOL(i,j,t,s) =e= cos(hh(s))*PL(i,j,t)+sin(hh(s))*QL(i,j,t);

Voltage_(i,t).. voltage(i,t) =e= v_min + sum(k, dV(i,t,k));



Z_decentral_..  z_dc =e= sum((i,t), price_of_power(i,t) * Pald(i,t));

Pald_(i,t).. Pald(i,t) =e= sum(a, PCh(a, t) * is_connected_to(a, i, t));
Qald_(i,t).. Qald(i,t) =e= 0;

SOC_initial_(a,t)$(ord(t) = 1).. SOC(a,t) =e= initial_SOC(a) + PCh(a,t)/capacity(a) - PCo(a,t)/capacity(a);
SOC_(a,t)$(ord(t) > 1)..         SOC(a,t) =e= SOC(a,t-1) + PCh(a,t)/capacity(a) - PCo(a,t)/capacity(a);

SOC_min_(a,t).. SOC(a,t) =g= SOC_min;
SOC_max_(a,t).. SOC(a,t) =l= 1.0;
  
eq_soc_(a, t)$(ord(t) = tEnd).. SOC(a, t) =g= initial_SOC(a);

PCh_(a, t).. PCh(a, t) =e= charging_rate(a) * is_charging(a, t) * sum(i, is_connected_to(a, i, t));
PCo_(a, t).. PCo(a, t) =e=  discharge_rate(a) * (1-sum(i, is_connected_to(a,i,t)));


eq_34_.. P_34 =e= sum((a,t), P_34_time(a,t));

eq_34_t_(a, t,iterations)$(ord(iterations) = iteration).. P_34_time(a, t) =e= sum(i, price_of_power(i,t) * PCh(a,t) * is_connected_to(a, i, t)) - sum(i, lambda(t,iterations-1) * PCh(a, t) * is_connected_to(a, i, t));

eq_34_t_balance_(a, t,iterations)$(ord(iterations) = iteration).. P_34_time(a, t) =e= sum(i, price_of_power(i,t) * PCh(a,t) * is_connected_to(a,i,t)) - sum(i, lambda(t,iterations-1) * PCh(a,t) * is_connected_to(a,i,t));


eq_35_.. P_35 =e= sum(t, P_35_time(t));

eq_35_t_(t,iterations)$(ord(iterations) = iteration).. P_35_time(t) =e= sum(i, sqr(nG(t) * Pg(i,t))) +  sum(i, lambda(t,iterations-1) * (Pg(i,t) - Pbld(i,t) - sum(j, AL(i,j) * PL(i,j,t))));

eq_35_t_balance_(t,iterations)$(ord(iterations) = iteration).. P_35_time(t) =e= sum(i, sqr(Pg(i,t))) + sum(i, lambda(t,iterations-1) * Pg(i,t)) ;



Model dumb /Z_central_dumb_,HPF1_, HPF2_, HPF3_l_, HPF4_l_, SOL_v_, SOL_1_, SOL_2_, Voltage_, Pald_, Qald_, PCh_, PCo_, SOC_initial_, SOC_, SOC_min_, SOC_max_, eq_soc_/;


Model fully_central /Z_central_,HPF1_, HPF2_, HPF3_l_, HPF4_l_, SOL_v_, SOL_1_, SOL_2_, Voltage_, Z_decentral_, Pald_, Qald_, PCh_, PCo_, SOC_initial_, SOC_, SOC_min_, SOC_max_, eq_soc_/;


Model central_dual /HPF2_, HPF3_l_, HPF4_l_, SOL_v_, SOL_1_, SOL_2_, Voltage_, eq_35_, eq_35_t_/;

Model central_dual_balance /eq_35_, eq_35_t_balance_/;


Model decentral /Z_decentral_, Pald_, Qald_, PCh_, PCo_, SOC_initial_, SOC_, SOC_min_, SOC_max_, eq_soc_/;

Model decentral_dual /Pald_, Qald_, PCh_, PCo_, Z_decentral_, SOC_initial_, SOC_, SOC_min_, SOC_max_, eq_soc_, eq_34_, eq_34_t_/;

Model decentral_dual_balance /Pald_, PCh_, PCo_, Z_decentral_, SOC_initial_, SOC_, SOC_min_, SOC_max_, eq_soc_, eq_34_, eq_34_t_balance_/;



* Select the desired model by uncommenting it bellow. Or define a new one if you wish, your call. :)


$onText
* Dumb, reference

Solve dumb using lp maximizing z_c;

* Used for dats export:
*execute_unload "results.gdx", is_charging, Pg, Pald, SOC;
*execute "gdx2sqlite -i results.gdx -o results.db";

$offText

$onText
* Central

Solve fully_central using lp minimizing z_c;

* Used for dats export:
*execute_unload "results.gdx", is_charging, Pg, Pald, SOC;
*execute "gdx2sqlite -i results.gdx -o results.db";

$offText

$onText
* Decentral
Solve decentral using lp minimizing z_dc;

* Used for dats export:
*execute_unload "results.gdx", is_charging, Pg, Pald, SOC;
*execute "gdx2sqlite -i results.gdx -o results.db";

$offText


$onText
* Dual decomposiiton model

lambda(t, iterations) = 0;

Scalar alpha, bet, agent, bus, time;

for ( iteration = 1 to iteration_limit,

    Solve decentral_dual_balance using lp minimizing P_34;

    Solve central_dual_balance using nlp minimizing P_35;

    alpha = .5/iteration;

    for (time = 1 to tEnd,
        if (iteration > 1,             
            lambda(t, iterations)$(ord(t) = time and ord(iterations) = iteration) = lambda(t, iterations-1) + alpha * (sum(a, P_34_time.l(a,t)) - P_35_time.l(t));
        else 
            lambda(t, iterations)$(ord(t) = time and ord(iterations) = iteration) = 0;

        );
    );
);

display lambda, Pg.l, Pald.l, P_34_time.l, P_35_time.l, P_34.l, P_35.l, initial_SOC, is_charging.l, is_connected_to;

* Used for dats export:
*execute_unload "results.gdx", is_charging, Pg, Pald, SOC, lambda;
*execute "gdx2sqlite -i results.gdx -o results.db";

$offText


